/*
 *
 *   Copyright (c) 2015 Samuel Degrande
 *
 *  This file is part of Freedroid
 *
 *  Freedroid is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Freedroid is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Freedroid; see the file COPYING. If not, write to the
 *  Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *  MA  02111-1307  USA
 *
 */

#ifndef _codeset_h
#define _codeset_h

#include <stdint.h>
#include <libgen.h>

struct codeset_element {
	int empty;
	uint16_t code;
	char *name;
};

char *cs_available_codesets();
void cs_check_codeset(const char*);
int cs_code_is_empty(unsigned int, const char *);
char *cs_font_char_name(char *, char *, unsigned int, const char *);
unsigned int cs_name_to_code(const char *, const char *);
uint16_t cs_code_to_unicode(unsigned int, const char *);
char *cs_code_to_name(unsigned int, const char *);

#endif
